class ALPHA_MODE:
    UNKNOWN = 0
    PREMULTIPLIED = 1
    STRAIGHT = 2
    IGNORE = 3


class ANTIALIAS_MODE:
    PER_PRIMITIVE = 0
    ALIASED = 1


class ARC_SIZE:
    SMALL = 0
    LARGE = 1


class INTERPOLATION_MODE_DEFINITION:
    NEAREST_NEIGHBOR = 0
    LINEAR = 1
    CUBIC = 2
    MULTI_SAMPLE_LINEAR = 3
    ANISOTROPIC = 4
    HIGH_QUALITY_CUBIC = 5
    FANT = 6
    MIPMAP_LINEAR = 7


class BITMAP_INTERPOLATION_MODE:
    NEAREST_NEIGHBOR = 0
    LINEAR = 1


class CAP_STYLE:
    FLAT = 0
    SQUARE = 1
    ROUND = 2
    TRIANGLE = 3


class COMBINE_MODE:
    UNION = 0
    INTERSECT = 1
    XOR = 2
    EXCLUDE = 3


class COMPATIBLE_RENDER_TARGET_OPTIONS:
    NONE = 0x00000000
    GDI_COMPATIBLE = 0x00000001


class DASH_STYLE:
    SOLID = 0
    DASH = 1
    DOT = 2
    DASH_DOT = 3
    DASH_DOT_DOT = 4
    CUSTOM = 5


class DC_INITIALIZE_MODE:
    COPY = 0
    CLEAR = 1


class DEBUG_LEVEL:
    NONE = 0
    ERROR = 1
    WARNING = 2
    INFORMATION = 3


class DRAW_TEXT_OPTIONS:
    NONE = 0x00000000
    NO_SNAP = 0x00000001
    CLIP = 0x00000002
    ENABLE_COLOR_FONT = 0x00000004
    DISABLE_COLOR_BITMAP_SNAPPING = 0x00000008


class EXTEND_MODE:
    CLAMP = 0
    WRAP = 1
    MIRROR = 2


class D2D_FACTORY_TYPE:
    SINGLE_THREADED = 0
    MULTI_THREADED = 1


class FEATURE_LEVEL:
    DEFAULT = 0
    LEVEL_9 = 0x9100
    LEVEL_10 = 0xA000


class FIGURE_BEGIN:
    FILLED = 0
    HOLLOW = 1


class FIGURE_END:
    OPEN = 0
    CLOSED = 1


class FILL_MODE:
    ALTERNATE = 0
    WINDING = 1


class GAMMA:
    GAMMA_2_2 = 0
    GAMMA_1_0 = 1


class GEOMETRY_RELATION:
    UNKNOWN = 0
    DISJOINT = 1
    IS_CONTAINED = 2
    CONTAINS = 3
    OVERLAP = 4


class GEOMETRY_SIMPLIFICATION_OPTION:
    CUBICS_AND_LINES = 0
    OPTION_LINES = 1


class LAYER_OPTIONS:
    NONE = 0x00000000
    INITIALIZE_FOR_CLEARTYPE = 0x00000001


class LINE_JOIN:
    MITER = 0
    BEVEL = 1
    ROUND = 2
    MITER_OR_BEVEL = 3


class MEASURING_MODE:
    NATURAL = 0
    GDI_CLASSIC = 1
    GDI_NATURAL = 2


class OPACITY_MASK_CONTENT:
    GRAPHICS = 0
    TEXT_NATURAL = 1
    TEXT_GDI_COMPATIBLE = 2


class PATH_SEGMENT:
    NONE = 0x00000000
    FORCE_UNSTROKED = 0x00000001
    FORCE_ROUND_LINE_JOIN = 0x00000002


class PRESENT_OPTIONS:
    NONE = 0x00000000
    RETAIN_CONTENTS = 0x00000001
    IMMEDIATELY = 0x00000002


class RENDER_TARGET_TYPE:
    DEFAULT = 0
    SOFTWARE = 1
    HARDWARE = 2


class RENDER_TARGET_USAGE:
    NONE = 0x00000000
    FORCE_BITMAP_REMOTING = 0x00000001
    GDI_COMPATIBLE = 0x00000002


class SWEEP_DIRECTION:
    COUNTER_CLOCKWISE = 0
    CLOCKWISE = 1


class TEXT_ANTIALIAS_MODE:
    DEFAULT = 0
    CLEARTYPE = 1
    GRAYSCALE = 2
    ALIASED = 3


class WINDOW_STATE:
    NONE = 0x00000000
    OCCLUDED = 0x00000001


class DWRITE_FACTORY_TYPE:
    SHARED = 0
    ISOLATED = 1


class FONT_FILE_TYPE:
    UNKNOWN = 0
    CFF = 1
    TRUETYPE = 2
    OPENTYPE_COLLECTION = 3
    TYPE1_PFM = 4
    TYPE1_PFB = 5
    VECTOR = 6
    BITMAP = 7
    TRUETYPE_COLLECTION = OPENTYPE_COLLECTION


class FONT_FACE_TYPE:
    CFF = 0
    TRUETYPE = 1
    OPENTYPE_COLLECTION = 2
    TYPE1 = 3
    VECTOR = 4
    BITMAP = 5
    UNKNOWN = 6
    RAW_CFF = 7
    TRUETYPE_COLLECTION = OPENTYPE_COLLECTION


class FONT_WEIGHT:
    THIN = 100
    EXTRA_LIGHT = 200
    ULTRA_LIGHT = 200
    LIGHT = 300
    SEMI_LIGHT = 350
    NORMAL = 400
    REGULAR = 400
    MEDIUM = 500
    DEMI_BOLD = 600
    BOLD = 600
    BOLD = 700
    EXTRA_BOLD = 800
    ULTRA_BOLD = 800
    BLACK = 900
    HEAVY = 900
    EXTRA_BLACK = 950
    ULTRA_BLACK = 950


class FONT_STRETCH:
    UNDEFINED = 0
    ULTRA_CONDENSED = 1
    EXTRA_CONDENSED = 2
    CONDENSED = 3
    SEMI_CONDENSED = 4
    NORMAL = 5
    MEDIUM = 5
    SEMI_EXPANDED = 6
    EXPANDED = 7
    EXTRA_EXPANDED = 8
    ULTRA_EXPANDED = 9


class FONT_STYLE:
    NORMAL = 0
    OBLIQUE = 1
    ITALIC = 2


class INFORMATIONAL_STRING_ID:
    NONE = 0
    COPYRIGHT_NOTICE = 1
    VERSION_STRINGS = 2
    TRADEMARK = 3
    MANUFACTURER = 4
    DESIGNER = 5
    DESIGNER_URL = 6
    DESCRIPTION = 7
    FONT_VENDOR_URL = 8
    LICENSE_DESCRIPTION = 9
    LICENSE_INFO_URL = 10
    WIN32_FAMILY_NAMES = 11
    WIN32_SUBFAMILY_NAMES = 12
    TYPOGRAPHIC_FAMILY_NAMES = 13
    TYPOGRAPHIC_SUBFAMILY_NAMES = 14
    SAMPLE_TEXT = 15
    FULL_NAME = 16
    POSTSCRIPT_NAME = 17
    POSTSCRIPT_CID_NAME = 18
    WEIGHT_STRETCH_STYLE_FAMILY_NAME = 19
    DESIGN_SCRIPT_LANGUAGE_TAG = 20
    SUPPORTED_SCRIPT_LANGUAGE_TAG = 21
    PREFERRED_FAMILY_NAMES = TYPOGRAPHIC_FAMILY_NAMES
    PREFERRED_SUBFAMILY_NAMES = TYPOGRAPHIC_SUBFAMILY_NAMES
    WWS_FAMILY_NAME = WEIGHT_STRETCH_STYLE_FAMILY_NAME


class FONT_SIMULATIONS:
    NONE = 0
    BOLD = 1
    OBLIQUE = 2


class PIXEL_GEOMETRY:
    FLAT = 0
    RGB = 1
    BGR = 2


class RENDERING_MODE:
    DEFAULT = 0
    ALIASED = 1
    GDI_CLASSIC = 2
    GDI_NATURAL = 3
    NATURAL = 4
    NATURAL_SYMMETRIC = 5
    OUTLINE = 6
    CLEARTYPE_GDI_CLASSIC = GDI_CLASSIC
    CLEARTYPE_GDI_NATURAL = GDI_NATURAL
    CLEARTYPE_NATURAL = NATURAL
    CLEARTYPE_NATURAL_SYMMETRIC = NATURAL_SYMMETRIC


class TEXT_ALIGNMENT:
    LEADING = 0
    TRAILING = 1
    CENTER = 2
    JUSTIFIED = 3


class PARAGRAPH_ALIGNMENT:
    NEAR = 0
    FAR = 1
    CENTER = 2


class WORD_WRAPPING:
    WRAP = 0
    NO_WRAP = 1
    EMERGENCY_BREAK = 2
    WHOLE_WORD = 3
    CHARACTER = 4


class READING_DIRECTION:
    LEFT_TO_RIGHT = 0
    RIGHT_TO_LEFT = 1
    TOP_TO_BOTTOM = 2
    BOTTOM_TO_TOP = 3


class FLOW_DIRECTION:
    TOP_TO_BOTTOM = 0
    BOTTOM_TO_TOP = 1
    LEFT_TO_RIGHT = 2
    RIGHT_TO_LEFT = 3


class TRIMMING_GRANULARITY:
    NONE = 0
    CHARACTER = 1
    WORD = 2


class BREAK_CONDITION:
    NEUTRAL = 0
    CAN_BREAK = 1
    MAY_NOT_BREAK = 2
    MUST_BREAK = 3


class LINE_SPACING_METHOD:
    DEFAULT = 0
    UNIFORM = 1
    PROPORTIONAL = 2


class FONT_FEATURE_TAG:
    ALTERNATIVE_FRACTIONS = 0x63726661
    PETITE_CAPITALS_FROM_CAPITALS = 0x63703263
    SMALL_CAPITALS_FROM_CAPITALS = 0x63733263
    CONTEXTUAL_ALTERNATES = 0x746C6163
    CASE_SENSITIVE_FORMS = 0x65736163
    GLYPH_COMPOSITION_DECOMPOSITION = 0x706D6363
    CONTEXTUAL_LIGATURES = 0x67696C63
    CAPITAL_SPACING = 0x70737063
    CONTEXTUAL_SWASH = 0x68777363
    CURSIVE_POSITIONING = 0x73727563
    DEFAULT = 0x746C6664
    DISCRETIONARY_LIGATURES = 0x67696C64
    EXPERT_FORMS = 0x74707865
    FRACTIONS = 0x63617266
    FULL_WIDTH = 0x64697766
    HALF_FORMS = 0x666C6168
    HALANT_FORMS = 0x6E6C6168
    ALTERNATE_HALF_WIDTH = 0x746C6168
    HISTORICAL_FORMS = 0x74736968
    HORIZONTAL_KANA_ALTERNATES = 0x616E6B68
    HISTORICAL_LIGATURES = 0x67696C68
    HALF_WIDTH = 0x64697768
    HOJO_KANJI_FORMS = 0x6F6A6F68
    JIS04_FORMS = 0x3430706A
    JIS78_FORMS = 0x3837706A
    JIS83_FORMS = 0x3338706A
    JIS90_FORMS = 0x3039706A
    KERNING = 0x6E72656B
    STANDARD_LIGATURES = 0x6167696C
    LINING_FIGURES = 0x6D756E6C
    LOCALIZED_FORMS = 0x6C636F6C
    MARK_POSITIONING = 0x6B72616D
    MATHEMATICAL_GREEK = 0x6B72676D
    MARK_TO_MARK_POSITIONING = 0x6B6D6B6D
    ALTERNATE_ANNOTATION_FORMS = 0x746C616E
    NLC_KANJI_FORMS = 0x6B636C6E
    OLD_STYLE_FIGURES = 0x6D756E6F
    ORDINALS = 0x6E64726F
    PROPORTIONAL_ALTERNATE_WIDTH = 0x746C6170
    PETITE_CAPITALS = 0x70616370
    PROPORTIONAL_FIGURES = 0x6D756E70
    PROPORTIONAL_WIDTHS = 0x64697770
    QUARTER_WIDTHS = 0x64697771
    REQUIRED_LIGATURES = 0x67696C72
    RUBY_NOTATION_FORMS = 0x79627572
    STYLISTIC_ALTERNATES = 0x746C6173
    SCIENTIFIC_INFERIORS = 0x666E6973
    SMALL_CAPITALS = 0x70636D73
    SIMPLIFIED_FORMS = 0x6C706D73
    STYLISTIC_SET_1 = 0x31307373
    STYLISTIC_SET_2 = 0x32307373
    STYLISTIC_SET_3 = 0x33307373
    STYLISTIC_SET_4 = 0x34307373
    STYLISTIC_SET_5 = 0x35307373
    STYLISTIC_SET_6 = 0x36307373
    STYLISTIC_SET_7 = 0x37307373
    STYLISTIC_SET_8 = 0x38307373
    STYLISTIC_SET_9 = 0x39307373
    STYLISTIC_SET_10 = 0x30317373
    STYLISTIC_SET_11 = 0x31317373
    STYLISTIC_SET_12 = 0x32317373
    STYLISTIC_SET_13 = 0x33317373
    STYLISTIC_SET_14 = 0x34317373
    STYLISTIC_SET_15 = 0x35317373
    STYLISTIC_SET_16 = 0x36317373
    STYLISTIC_SET_17 = 0x37317373
    STYLISTIC_SET_18 = 0x38317373
    STYLISTIC_SET_19 = 0x39317373
    STYLISTIC_SET_20 = 0x30327373
    SUBSCRIPT = 0x73627573
    SUPERSCRIPT = 0x73707573
    SWASH = 0x68737773
    TITLING = 0x6C746974
    TRADITIONAL_NAME_FORMS = 0x6D616E74
    TABULAR_FIGURES = 0x6D756E74
    TRADITIONAL_FORMS = 0x64617274
    THIRD_WIDTHS = 0x64697774
    UNICASE = 0x63696E75
    VERTICAL_WRITING = 0x74726576
    VERTICAL_ALTERNATES_AND_ROTATION = 0x32747276
    SLASHED_ZERO = 0x6F72657A


class SCRIPT_SHAPES:
    DEFAULT = 0
    NO_VISUAL = 1


class NUMBER_SUBSTITUTION_METHOD:
    FROM_CULTURE = 0
    CONTEXTUAL = 1
    NONE = 2
    NATIONAL = 3
    TRADITIONAL = 4


ALPHA_MAX = 255


class TEXTURE_TYPE:
    ALIASED_1x1 = 0
    CLEARTYPE_3x1 = 1


class DXGI_FORMAT:
    UNKNOWN = 0
    R32G32B32A32_TYPELESS = 1
    R32G32B32A32_FLOAT = 2
    R32G32B32A32_UINT = 3
    R32G32B32A32_SINT = 4
    R32G32B32_TYPELESS = 5
    R32G32B32_FLOAT = 6
    R32G32B32_UINT = 7
    R32G32B32_SINT = 8
    R16G16B16A16_TYPELESS = 9
    R16G16B16A16_FLOAT = 10
    R16G16B16A16_UNORM = 11
    R16G16B16A16_UINT = 12
    R16G16B16A16_SNORM = 13
    R16G16B16A16_SINT = 14
    R32G32_TYPELESS = 15
    R32G32_FLOAT = 16
    R32G32_UINT = 17
    R32G32_SINT = 18
    R32G8X24_TYPELESS = 19
    D32_FLOAT_S8X24_UINT = 20
    R32_FLOAT_X8X24_TYPELESS = 21
    X32_TYPELESS_G8X24_UINT = 22
    R10G10B10A2_TYPELESS = 23
    R10G10B10A2_UNORM = 24
    R10G10B10A2_UINT = 25
    R11G11B10_FLOAT = 26
    R8G8B8A8_TYPELESS = 27
    R8G8B8A8_UNORM = 28
    R8G8B8A8_UNORM_SRGB = 29
    R8G8B8A8_UINT = 30
    R8G8B8A8_SNORM = 31
    R8G8B8A8_SINT = 32
    R16G16_TYPELESS = 33
    R16G16_FLOAT = 34
    R16G16_UNORM = 35
    R16G16_UINT = 36
    R16G16_SNORM = 37
    R16G16_SINT = 38
    R32_TYPELESS = 39
    D32_FLOAT = 40
    R32_FLOAT = 41
    R32_UINT = 42
    R32_SINT = 43
    R24G8_TYPELESS = 44
    D24_UNORM_S8_UINT = 45
    R24_UNORM_X8_TYPELESS = 46
    X24_TYPELESS_G8_UINT = 47
    R8G8_TYPELESS = 48
    R8G8_UNORM = 49
    R8G8_UINT = 50
    R8G8_SNORM = 51
    R8G8_SINT = 52
    R16_TYPELESS = 53
    R16_FLOAT = 54
    D16_UNORM = 55
    R16_UNORM = 56
    R16_UINT = 57
    R16_SNORM = 58
    R16_SINT = 59
    R8_TYPELESS = 60
    R8_UNORM = 61
    R8_UINT = 62
    R8_SNORM = 63
    R8_SINT = 64
    A8_UNORM = 65
    R1_UNORM = 66
    R9G9B9E5_SHAREDEXP = 67
    R8G8_B8G8_UNORM = 68
    G8R8_G8B8_UNORM = 69
    BC1_TYPELESS = 70
    BC1_UNORM = 71
    BC1_UNORM_SRGB = 72
    BC2_TYPELESS = 73
    BC2_UNORM = 74
    BC2_UNORM_SRGB = 75
    BC3_TYPELESS = 76
    BC3_UNORM = 77
    BC3_UNORM_SRGB = 78
    BC4_TYPELESS = 79
    BC4_UNORM = 80
    BC4_SNORM = 81
    BC5_TYPELESS = 82
    BC5_UNORM = 83
    BC5_SNORM = 84
    B5G6R5_UNORM = 85
    B5G5R5A1_UNORM = 86
    B8G8R8A8_UNORM = 87
    B8G8R8X8_UNORM = 88
    R10G10B10_XR_BIAS_A2_UNORM = 89
    B8G8R8A8_TYPELESS = 90
    B8G8R8A8_UNORM_SRGB = 91
    B8G8R8X8_TYPELESS = 92
    B8G8R8X8_UNORM_SRGB = 93
    BC6H_TYPELESS = 94
    BC6H_UF16 = 95
    BC6H_SF16 = 96
    BC7_TYPELESS = 97
    BC7_UNORM = 98
    BC7_UNORM_SRGB = 99
    AYUV = 100
    Y410 = 101
    Y416 = 102
    NV12 = 103
    P010 = 104
    P016 = 105
    FORMAT_420_OPAQUE = 106
    YUY2 = 107
    Y210 = 108
    Y216 = 109
    NV11 = 110
    AI44 = 111
    IA44 = 112
    P8 = 113
    A8P8 = 114
    B4G4R4A4_UNORM = 115
    P208 = 130
    V208 = 131
    V408 = 132
    SAMPLER_FEEDBACK_MIN_MIP_OPAQUE = 189
    SAMPLER_FEEDBACK_MIP_REGION_USED_OPAQUE = 190



# cython: language_level=3

include "_pyd2d_const.pyi"
from libc.stdint cimport int32_t, uint32_t, uint64_t, intptr_t
from libc.stddef cimport wchar_t
from cpython.mem cimport PyMem_Free

cdef extern from *:
    """
    #define UNICODE
    #define _UNICODE
    """
    wchar_t* PyUnicode_AsWideCharString(object unicode, Py_ssize_t *size)
    object PyUnicode_FromWideChar(wchar_t *wstr, Py_ssize_t size)

cdef extern from "windows.h":
    ctypedef int32_t HRESULT
    ctypedef uint32_t ULONG
    ctypedef void* HWND
    ctypedef uint32_t UINT32
    ctypedef float FLOAT
    ctypedef uint64_t UINT64
    ctypedef uint32_t UINT
    ctypedef wchar_t WCHAR

    ctypedef struct GUID:
        unsigned long Data1
        unsigned short Data2
        unsigned short Data3
        unsigned char Data4[8]

    unsigned int FormatMessageW(
        unsigned long dwFlags,
        void *lpSource,
        unsigned long dwMessageId,
        unsigned long dwLanguageId,
        wchar_t* lpBuffer,
        unsigned long nSize,
        void *Arguments)
    bint SUCCEEDED(HRESULT hr)
    bint FAILED(HRESULT hr)

cdef extern from "objbase.h":
    HRESULT CoInitializeEx(
        void* pvReserved,
        unsigned long dwCoInit)
    
    cdef cppclass IUnknown:
        void Release()

cdef extern from "dwrite.h":
    """
    #define IID_IDWriteFactory __uuidof(IDWriteFactory)
    """
    ctypedef int DWRITE_FACTORY_TYPE
    ctypedef int DWRITE_FONT_STYLE
    ctypedef int DWRITE_FONT_WEIGHT
    ctypedef int DWRITE_FONT_STRETCH
    
    ctypedef struct DWRITE_TEXT_METRICS:
        FLOAT  left
        FLOAT  top
        FLOAT  width
        FLOAT  widthIncludingTrailingWhitespace
        FLOAT  height
        FLOAT  layoutWidth
        FLOAT  layoutHeight
        UINT32 maxBidiReorderingDepth
        UINT32 lineCount

    cdef GUID IID_IDWriteFactory

    cdef cppclass IDWriteTextLayout:
        FLOAT GetMaxWidth()
        HRESULT GetMetrics(DWRITE_TEXT_METRICS *textMetrics)
    cdef cppclass IDWriteFactory:
        HRESULT CreateTextFormat(
            const WCHAR *family_name,
            IDWriteFontCollection *collection,
            DWRITE_FONT_WEIGHT weight,
            DWRITE_FONT_STYLE style,
            DWRITE_FONT_STRETCH stretch,
            FLOAT size,
            const WCHAR *locale,
            IDWriteTextFormat **format)
        HRESULT CreateTextLayout(
            WCHAR *string,
            UINT32 stringLength,
            IDWriteTextFormat *textFormat,
            FLOAT maxWidth,
            FLOAT maxHeight,
            IDWriteTextLayout **textLayout
        )
    cdef cppclass IDWriteFontCollection
    cdef cppclass IDWriteTextFormat

    HRESULT DWriteCreateFactory(DWRITE_FACTORY_TYPE factoryType, const GUID& iid, IUnknown** factory)


cdef extern from "d2d1.h":
    ctypedef int D2D1_DEBUG_LEVEL
    ctypedef int D2D1_FACTORY_TYPE
    ctypedef int D2D1_RENDER_TARGET_TYPE
    ctypedef int DXGI_FORMAT
    ctypedef int D2D1_ALPHA_MODE
    ctypedef int D2D1_RENDER_TARGET_USAGE
    ctypedef int D2D1_FEATURE_LEVEL
    ctypedef int D2D1_PRESENT_OPTIONS
    ctypedef int D2D1_CAP_STYLE
    ctypedef int D2D1_LINE_JOIN
    ctypedef int D2D1_DASH_STYLE
    ctypedef int D2D1_SWEEP_DIRECTION
    ctypedef int D2D1_ARC_SIZE
    ctypedef int D2D1_FILL_MODE
    ctypedef int D2D1_ANTIALIAS_MODE
    ctypedef int D2D1_BITMAP_INTERPOLATION_MODE
    ctypedef int D2D1_DRAW_TEXT_OPTIONS
    ctypedef int DWRITE_MEASURING_MODE
    ctypedef int D2D1_FIGURE_BEGIN
    ctypedef int D2D1_FIGURE_END
    
    ctypedef struct D2D1_FACTORY_OPTIONS:
        D2D1_DEBUG_LEVEL debugLevel
    ctypedef struct D2D1_PIXEL_FORMAT:
         DXGI_FORMAT     format
         D2D1_ALPHA_MODE alphaMode
    ctypedef struct D2D1_RENDER_TARGET_PROPERTIES:
         D2D1_RENDER_TARGET_TYPE  type
         D2D1_PIXEL_FORMAT        pixelFormat
         float                    dpiX
         float                    dpiY
         D2D1_RENDER_TARGET_USAGE usage
         D2D1_FEATURE_LEVEL       minLevel
    ctypedef struct D2D1_SIZE_U:
        UINT32 width
        UINT32 height
    ctypedef struct D2D1_HWND_RENDER_TARGET_PROPERTIES:
       HWND                 hwnd
       D2D1_SIZE_U          pixelSize
       D2D1_PRESENT_OPTIONS presentOptions
    ctypedef struct D3DCOLORVALUE:
        float r
        float g
        float b
        float a
    ctypedef D3DCOLORVALUE D2D1_COLOR_F
    ctypedef struct D2D1_MATRIX_3X2_F:
        float m11
        float m12
        float m21
        float m22
        float dx
        float dy
    ctypedef struct D2D1_BRUSH_PROPERTIES:
        FLOAT             opacity
        D2D1_MATRIX_3X2_F transform
    ctypedef struct D2D1_RECT_F:
        float left
        float top
        float right
        float bottom
    ctypedef UINT64 D2D1_TAG
    ctypedef struct D2D1_STROKE_STYLE_PROPERTIES:
        D2D1_CAP_STYLE  startCap
        D2D1_CAP_STYLE  endCap
        D2D1_CAP_STYLE  dashCap
        D2D1_LINE_JOIN  lineJoin
        FLOAT           miterLimit
        D2D1_DASH_STYLE dashStyle
        FLOAT           dashOffset
    ctypedef struct D2D1_POINT_2F:
        float x
        float y
    ctypedef struct D2D1_BEZIER_SEGMENT:
        D2D1_POINT_2F point1
        D2D1_POINT_2F point2
        D2D1_POINT_2F point3
    ctypedef struct D2D1_SIZE_F:
        float width
        float height
    ctypedef struct D2D1_ARC_SEGMENT:
        D2D1_POINT_2F        point
        D2D1_SIZE_F          size
        FLOAT                rotationAngle
        D2D1_SWEEP_DIRECTION sweepDirection
        D2D1_ARC_SIZE        arcSize
    ctypedef struct D2D1_QUADRATIC_BEZIER_SEGMENT:
        D2D1_POINT_2F point1
        D2D1_POINT_2F point2
    ctypedef struct D2D1_ELLIPSE:
        D2D1_POINT_2F point
        FLOAT         radiusX
        FLOAT         radiusY
    ctypedef struct D2D1_BITMAP_PROPERTIES:
        D2D1_PIXEL_FORMAT pixelFormat
        FLOAT             dpiX
        FLOAT             dpiY

    cdef cppclass ID2D1Bitmap
    cdef cppclass ID2D1Brush:
        FLOAT GetOpacity()
    cdef cppclass ID2D1PathGeometry:
        HRESULT Open(ID2D1GeometrySink **geometrySink)
    cdef cppclass ID2D1StrokeStyle
    cdef cppclass ID2D1Factory:
        HRESULT CreateHwndRenderTarget(
            const D2D1_RENDER_TARGET_PROPERTIES *renderTargetProperties,
            const D2D1_HWND_RENDER_TARGET_PROPERTIES *hwndRenderTargetProperties,
            ID2D1HwndRenderTarget **hwndRenderTarget)
        HRESULT CreatePathGeometry(ID2D1PathGeometry **pathGeometry)
        HRESULT CreateStrokeStyle(
            const D2D1_STROKE_STYLE_PROPERTIES *strokeStyleProperties,
            const FLOAT *dashes,
            UINT dashesCount,
            ID2D1StrokeStyle **strokeStyle)
    cdef cppclass ID2D1Geometry
    cdef cppclass ID2D1GeometrySink:
        void AddArc(const D2D1_ARC_SEGMENT *arc)
        void AddBezier(const D2D1_BEZIER_SEGMENT *bezier)
        void AddLine(D2D1_POINT_2F point)
        void AddQuadraticBezier(const D2D1_QUADRATIC_BEZIER_SEGMENT *bezier)
        void BeginFigure(D2D1_POINT_2F startPoint, D2D1_FIGURE_BEGIN figureBegin)
        HRESULT Close()
        void EndFigure(D2D1_FIGURE_END figureEnd)
        void SetFillMode(D2D1_FILL_MODE fillMode)
    cdef cppclass ID2D1HwndRenderTarget:
        HRESULT Resize(const D2D1_SIZE_U *pixelSize)
    cdef cppclass ID2D1RenderTarget:
        void BeginDraw()
        void Clear(const D2D1_COLOR_F *clearColor)
        HRESULT CreateBitmapFromWicBitmap(
            IWICBitmapSource *wicBitmapSource,
            const D2D1_BITMAP_PROPERTIES *bitmapProperties,
            ID2D1Bitmap **bitmap)
        HRESULT CreateSolidColorBrush(
            const D2D1_COLOR_F *color,
            const D2D1_BRUSH_PROPERTIES *brushProperties,
            ID2D1SolidColorBrush **solidColorBrush)
        void DrawBitmap(
            ID2D1Bitmap *bitmap,
            const D2D1_RECT_F& destinationRectangle,
            FLOAT opacity,
            D2D1_BITMAP_INTERPOLATION_MODE interpolationMode,
            const D2D1_RECT_F *sourceRectangle)
        void DrawEllipse(
            const D2D1_ELLIPSE *ellipse,
            ID2D1Brush *brush,
            FLOAT strokeWidth,
            ID2D1StrokeStyle *strokeStyle)
        void DrawGeometry(
            ID2D1Geometry *geometry,
            ID2D1Brush *brush,
            FLOAT strokeWidth,
            ID2D1StrokeStyle *strokeStyle)
        void DrawLine(
            D2D1_POINT_2F point0,
            D2D1_POINT_2F point1,
            ID2D1Brush *brush,
            FLOAT strokeWidth,
            ID2D1StrokeStyle *strokeStyle)
        void DrawRectangle(
            const D2D1_RECT_F *rect,
            ID2D1Brush *brush,
            FLOAT strokeWidth,
            ID2D1StrokeStyle *strokeStyle)
        void DrawTextW(
            const WCHAR *string,
            UINT stringLength,
            IDWriteTextFormat *textFormat,
            const D2D1_RECT_F &layoutRect,
            ID2D1Brush *defaultForegroundBrush,
            D2D1_DRAW_TEXT_OPTIONS options,
            DWRITE_MEASURING_MODE measuringMode)
        void DrawTextLayout(
            D2D1_POINT_2F origin,
            IDWriteTextLayout *textLayout,
            ID2D1Brush *defaultForegroundBrush,
            D2D1_DRAW_TEXT_OPTIONS options)
        HRESULT EndDraw(D2D1_TAG *tag1, D2D1_TAG *tag2)
        void FillEllipse(const D2D1_ELLIPSE *ellipse, ID2D1Brush *brush)
        void FillGeometry(
            ID2D1Geometry *geometry,
            ID2D1Brush *brush,
            ID2D1Brush *opacityBrush)
        void FillRectangle(const D2D1_RECT_F *rect, ID2D1Brush *brush)
        void GetTransform(D2D1_MATRIX_3X2_F *transform)
        void SetAntialiasMode(D2D1_ANTIALIAS_MODE antialiasMode)
        void SetTransform(const D2D1_MATRIX_3X2_F *transform)
    cdef cppclass ID2D1SimplifiedGeometrySink:
        void BeginFigure(D2D1_POINT_2F startPoint, int figureBegin)
        HRESULT Close()
        void EndFigure(int figureEnd)
        void SetFillMode(D2D1_FILL_MODE fillMode)
    cdef cppclass ID2D1SolidColorBrush:
        void SetColor(const D2D1_COLOR_F *color)
    cdef cppclass IWICBitmapSource

    cdef const GUID IID_ID2D1Factory

    HRESULT D2D1CreateFactory(
        D2D1_FACTORY_TYPE factoryType,
        const GUID& riid,
        const D2D1_FACTORY_OPTIONS *pFactoryOptions,
        void **ppIFactory)


cdef getHRESULTstring(int i):
    cdef int strBufLen = 0x1000
    cdef wchar_t[0x1000] strBuf
    cdef unsigned int ncopied = FormatMessageW(
        0x1200,
        NULL,
        i,
        0,
        strBuf,
        strBufLen,
        NULL)
    if ncopied == 0:
        return 'unknown error 0x%08x' % i
    return PyUnicode_FromWideChar(strBuf, ncopied).rstrip("\r\n")


class COMError(OSError):
    def __init__(self, hresult):
        self.hresult = hresult
        self.args = (getHRESULTstring(hresult),)


class Direct2DError(COMError):
    pass

def InitializeCOM(options=0):
    cdef HRESULT hr = CoInitializeEx(NULL, options)
    if hr == 1:
        # already initialized
        pass
    elif FAILED(hr):
        raise COMError(hr)


cdef class COMObject:
    cdef void* ptr

    def __init__(self, int ptr):
        self.ptr = <void*><intptr_t>ptr

    def __del__(self):
        try:
            (<IUnknown*>self.ptr).Release()
        except AttributeError:
            pass


_d2d_factory = None

def GetD2DFactory():
    global _d2d_factory
    if _d2d_factory is None:
        _d2d_factory = D2DFactory()
    return _d2d_factory

from libc.stdio cimport printf

cdef class D2DFactory(COMObject):
    def __init__(self, int factoryType=0, int debugLevel=0):
        cdef D2D1_FACTORY_OPTIONS options
        cdef ID2D1Factory* factory
        options.debugLevel = <D2D1_DEBUG_LEVEL>debugLevel
        res = D2D1CreateFactory(<D2D1_FACTORY_TYPE>factoryType, IID_ID2D1Factory, &options, <void**>&factory)
        if FAILED(res):
            raise Direct2DError(res)
        COMObject.__init__(self, <int><intptr_t>factory)

    def CreateHwndRenderTarget(
            self,
            intptr_t hwnd,
            int width,
            int height,
            int presentOptions=0,
            int rtType=0,
            int pixelFormat=0,
            int alphaMode=0,
            float dpiX=0,
            float dpiY=0,
            int usage=0,
            int featureLevel=0):
        cdef D2D1_RENDER_TARGET_PROPERTIES rtp
        rtp.type = <D2D1_RENDER_TARGET_TYPE>rtType
        rtp.pixelFormat.format = <DXGI_FORMAT>pixelFormat
        rtp.pixelFormat.alphaMode = <D2D1_ALPHA_MODE>alphaMode
        rtp.dpiX = dpiX
        rtp.dpiY = dpiY
        rtp.usage = <D2D1_RENDER_TARGET_USAGE>usage
        rtp.minLevel = <D2D1_FEATURE_LEVEL>featureLevel
        cdef D2D1_HWND_RENDER_TARGET_PROPERTIES hrtp
        hrtp.hwnd = <HWND>hwnd
        hrtp.pixelSize.width = width
        hrtp.pixelSize.height = height
        hrtp.presentOptions = <D2D1_PRESENT_OPTIONS>presentOptions
        cdef ID2D1HwndRenderTarget* target
        res = (<ID2D1Factory*>self.ptr).CreateHwndRenderTarget(&rtp, &hrtp, <ID2D1HwndRenderTarget**>(&target))
        if FAILED(res):
            raise Direct2DError(res)
        return HWNDRenderTarget(<int><intptr_t>target)

    def CreatePathGeometry(self):
        cdef ID2D1PathGeometry* pgm
        res = (<ID2D1Factory*>self.ptr).CreatePathGeometry(<ID2D1PathGeometry**>&pgm)
        if FAILED(res):
            raise Direct2DError(res)
        return PathGeometry(<int><intptr_t>pgm)

    def CreateStrokeStyle(
            self,
            int startCap=0,
            int endCap=0,
            int dashCap=0,
            int lineJoin=0,
            float miterLimit=10.0,
            int dashStyle=0,
            float dashOffset=0.0):
        cdef D2D1_STROKE_STYLE_PROPERTIES ssp
        ssp.startCap = <D2D1_CAP_STYLE>startCap
        ssp.endCap = <D2D1_CAP_STYLE>endCap
        ssp.dashCap = <D2D1_CAP_STYLE>dashCap
        ssp.lineJoin = <D2D1_LINE_JOIN>lineJoin
        ssp.miterLimit = miterLimit
        ssp.dashStyle = <D2D1_DASH_STYLE>dashStyle
        ssp.dashOffset = dashOffset
        cdef ID2D1StrokeStyle *sstyle
        res = (<ID2D1Factory*>self.ptr).CreateStrokeStyle(&ssp, NULL, 0, <ID2D1StrokeStyle**>&sstyle)
        if FAILED(res):
            raise Direct2DError(res)
        return StrokeStyle(<int><intptr_t>sstyle)


cdef class Resource(COMObject):
    pass


cdef class RenderTarget(Resource):
    def BeginDraw(self):
        (<ID2D1RenderTarget*>self.ptr).BeginDraw()

    def Clear(self, float r, float g, float b, float a=1.0):
        cdef D2D1_COLOR_F color
        color.r = r
        color.g = g
        color.b = b
        color.a = a
        (<ID2D1RenderTarget*>self.ptr).Clear(&color)

    # def CreateBitmapFromWicBitmap(self, source, int dxgiFormat=0, int alphaMode=0, float dpiX=0.0, float dpiY=0.0):
    #     cdef D2D1_BITMAP_PROPERTIES properties
    #     properties.pixelFormat.format = dxgiFormat
    #     properties.pixelFormat.alphaMode = alphaMode
    #     properties.dpiX = dpiX
    #     properties.dpiY = dpiY
    #     cdef ID2D1Bitmap *bitmap
    #     res = (<ID2D1RenderTarget*>self.ptr).CreateBitmapFromWicBitmap(
    #         <IWICBitmapSource*>source.ptr,
    #         &properties,
    #         &bitmap)
    #     if FAILED(res):
    #         raise Direct2DError(res)
    #     return Bitmap(<int>bitmap)

    def CreateSolidColorBrush(self, float r, float g, float b, float a=1.0, float opacity=1.0):
        cdef D2D1_COLOR_F color
        color.r = r
        color.g = g
        color.b = b
        color.a = a
        cdef D2D1_BRUSH_PROPERTIES bprop
        bprop.opacity = opacity
        cdef ID2D1SolidColorBrush *brush
        res = (<ID2D1RenderTarget*>self.ptr).CreateSolidColorBrush(&color, &bprop, &brush)
        if FAILED(res):
            raise Direct2DError(res)
        return SolidColorBrush(<int><intptr_t>brush)

    def DrawBitmap(
            self,
            Bitmap bitmap,
            float l,
            float t,
            float r,
            float b,
            float opacity=1.0,
            interpolationMode=1,
            srcRect=None):
        cdef D2D1_RECT_F dest, src
        dest.left = l
        dest.top = t
        dest.right = r
        dest.bottom = b
        cdef D2D1_RECT_F *srcRectPtr = NULL
        if srcRect is not None:
            src.left, src.top, src.right, src.bottom = srcRect
            srcRectPtr = &src
        (<ID2D1RenderTarget*>self.ptr).DrawBitmap(
            <ID2D1Bitmap*>bitmap.ptr,
            dest,
            opacity,
            <D2D1_BITMAP_INTERPOLATION_MODE>interpolationMode,
            srcRectPtr)

    def DrawEllipse(
            self,
            float cx,
            float cy,
            float rx,
            float ry,
            Brush brush,
            float strokeWidth=1.0,
            StrokeStyle strokeStyle=None):
        cdef D2D1_ELLIPSE el
        el.point.x = cx
        el.point.y = cy
        el.radiusX = rx
        el.radiusY = ry
        cdef ID2D1StrokeStyle *sstyle
        if strokeStyle is None:
            sstyle = NULL
        else:
            sstyle = <ID2D1StrokeStyle*>strokeStyle.ptr
        (<ID2D1RenderTarget*>self.ptr).DrawEllipse(&el, <ID2D1Brush*>brush.ptr, strokeWidth, sstyle)

    def DrawGeometry(self, Geometry geometry, Brush brush, float strokeWidth=1.0, StrokeStyle strokeStyle=None):
        cdef ID2D1StrokeStyle *sstyle
        if strokeStyle is None:
            sstyle = NULL
        else:
            sstyle = <ID2D1StrokeStyle*>strokeStyle.ptr
        (<ID2D1RenderTarget*>self.ptr).DrawGeometry(
            <ID2D1Geometry*>geometry.ptr,
            <ID2D1Brush*>brush.ptr,
            strokeWidth,
            sstyle)

    def DrawLine(
            self,
            float x1,
            float y1,
            float x2,
            float y2,
            Brush brush,
            float strokeWidth=1.0,
            StrokeStyle strokeStyle=None):
        cdef D2D1_POINT_2F point0
        point0.x = x1
        point0.y = y1
        cdef D2D1_POINT_2F point1
        point1.x = x2
        point1.y = y2
        cdef ID2D1StrokeStyle *sstyle
        if strokeStyle is None:
            sstyle = NULL
        else:
            sstyle = <ID2D1StrokeStyle*>strokeStyle.ptr
        (<ID2D1RenderTarget*>self.ptr).DrawLine(point0, point1, <ID2D1Brush*>brush.ptr, strokeWidth, sstyle)

    def DrawRectangle(
            self,
            float l,
            float t,
            float r,
            float b,
            Brush brush,
            float strokeWidth=1.0,
            StrokeStyle strokeStyle=None):
        cdef D2D1_RECT_F rect
        rect.left = l
        rect.top = t
        rect.right = r
        rect.bottom = b
        cdef ID2D1StrokeStyle *sstyle
        if strokeStyle is None:
            sstyle = NULL
        else:
            sstyle = <ID2D1StrokeStyle*>strokeStyle.ptr
        (<ID2D1RenderTarget*>self.ptr).DrawRectangle(&rect, <ID2D1Brush*>brush.ptr, strokeWidth, sstyle)

    def DrawText(
            self,
            str text,
            TextFormat textFormat,
            float l,
            float t,
            float r,
            float b,
            Brush brush,
            int options=0,
            int measuringMode=0):
        cdef D2D1_RECT_F rect
        rect.left = l
        rect.top = t
        rect.right = r
        rect.bottom = b
        cdef wchar_t *textBuf
        cdef Py_ssize_t _textLength
        textBuf = PyUnicode_AsWideCharString(text, &_textLength)
        cdef UINT textLength
        textLength = <UINT>_textLength
        if textBuf == NULL:
            raise MemoryError
        (<ID2D1RenderTarget*>self.ptr).DrawTextW(
            textBuf,
            <UINT>textLength,
            <IDWriteTextFormat*>textFormat.ptr,
            rect,
            <ID2D1Brush*>brush.ptr,
            <D2D1_DRAW_TEXT_OPTIONS>options,
            <DWRITE_MEASURING_MODE>measuringMode)
        PyMem_Free(<void*>textBuf)

    def DrawTextLayout(self, float x, float y, TextLayout textLayout, Brush brush, int options=0):
        cdef D2D1_POINT_2F pt
        pt.x = x
        pt.y = y
        (<ID2D1RenderTarget*>self.ptr).DrawTextLayout(
            pt,
            <IDWriteTextLayout*>textLayout.ptr,
            <ID2D1Brush*>brush.ptr,
            <D2D1_DRAW_TEXT_OPTIONS>options)

    def EndDraw(self):
        res = (<ID2D1RenderTarget*>self.ptr).EndDraw(NULL, NULL)
        if FAILED(res):
            raise Direct2DError(res)

    def FillEllipse(self, float cx, float cy, float rx, float ry, Brush brush):
        cdef D2D1_ELLIPSE el
        el.point.x = cx
        el.point.y = cy
        el.radiusX = rx
        el.radiusY = ry
        (<ID2D1RenderTarget*>self.ptr).FillEllipse(&el, <ID2D1Brush*>brush.ptr)

    def FillGeometry(self, Geometry geometry, Brush brush):
        (<ID2D1RenderTarget*>self.ptr).FillGeometry(<ID2D1Geometry*>geometry.ptr, <ID2D1Brush*>brush.ptr, NULL)

    def FillRectangle(self, float l, float t, float r, float b, Brush brush):
        cdef D2D1_RECT_F rect
        rect.left = l
        rect.top = t
        rect.right = r
        rect.bottom = b
        (<ID2D1RenderTarget*>self.ptr).FillRectangle(&rect, <ID2D1Brush*>brush.ptr)

    def GetTransform(self):
        cdef D2D1_MATRIX_3X2_F mat
        (<ID2D1RenderTarget*>self.ptr).GetTransform(&mat)
        return mat.m11, mat.m12, mat.m21, mat.m22, mat.dx, mat.dy

    def SetAntialiasMode(self, int mode):
        (<ID2D1RenderTarget*>self.ptr).SetAntialiasMode(<D2D1_ANTIALIAS_MODE>mode)

    def SetTransform(self, float m11=1, float m12=0, float m21=0, float m22=1, float dx=0, float dy=0):
        cdef D2D1_MATRIX_3X2_F mat
        mat.m11 = m11
        mat.m12 = m12
        mat.m21 = m21
        mat.m22 = m22
        mat.dx = dx
        mat.dy = dy
        (<ID2D1RenderTarget*>self.ptr).SetTransform(&mat)


cdef class HWNDRenderTarget(RenderTarget):
    def Resize(self, int width, int height):
        cdef D2D1_SIZE_U size
        size.width = width
        size.height = height
        res = (<ID2D1HwndRenderTarget*>self.ptr).Resize(&size)
        if FAILED(res):
            raise Direct2DError(res)


cdef class Brush(Resource):
    def GetOpacity(self):
        return (<ID2D1Brush*>self.ptr).GetOpacity()


cdef class SolidColorBrush(Brush):
    def SetColor(self, float r, float g, float b, float a=1.0):
        cdef D2D1_COLOR_F color
        color.r = r
        color.g = g
        color.b = b
        color.a = a
        (<ID2D1SolidColorBrush*>self.ptr).SetColor(&color)


cdef class StrokeStyle(Resource):
    pass


cdef class Geometry(Resource):
    pass


cdef class PathGeometry(Geometry):
    def Open(self):
        cdef ID2D1GeometrySink *gs
        res = (<ID2D1PathGeometry*>self.ptr).Open(&gs)
        if FAILED(res):
            raise Direct2DError(res)
        return GeometrySink(<int><intptr_t>gs)


cdef class SimplifiedGeometrySink(COMObject):
    def BeginFigure(self, float x, float y, int figureBegin=0):
        cdef D2D1_POINT_2F pt
        pt.x = x
        pt.y = y
        (<ID2D1SimplifiedGeometrySink*>self.ptr).BeginFigure(pt, <D2D1_FIGURE_BEGIN>figureBegin)

    def Close(self):
        res = (<ID2D1SimplifiedGeometrySink*>self.ptr).Close()
        if FAILED(res):
            raise Direct2DError(res)

    def EndFigure(self, int figureEnd=0):
        (<ID2D1SimplifiedGeometrySink*>self.ptr).EndFigure(<D2D1_FIGURE_END>figureEnd)

    def SetFillMode(self, int fillMode):
        (<ID2D1SimplifiedGeometrySink*>self.ptr).SetFillMode(<D2D1_FILL_MODE>fillMode)


cdef class GeometrySink(SimplifiedGeometrySink):
    def AddArc(self, float x, float y, float rx, float ry, float rotationAngle, int sweepDirection, int arcSize):
        cdef D2D1_ARC_SEGMENT arc
        arc.point.x = x
        arc.point.y = y
        arc.size.width = rx
        arc.size.height = ry
        arc.rotationAngle = rotationAngle
        arc.sweepDirection = <D2D1_SWEEP_DIRECTION>sweepDirection
        arc.arcSize = <D2D1_ARC_SIZE>arcSize
        (<ID2D1GeometrySink*>self.ptr).AddArc(&arc)

    def AddBezier(self, float x1, float y1, float x2, float y2, float x3, float y3):
        cdef D2D1_BEZIER_SEGMENT bz
        bz.point1.x = x1
        bz.point1.y = y1
        bz.point2.x = x2
        bz.point2.y = y2
        bz.point3.x = x3
        bz.point3.y = y3
        (<ID2D1GeometrySink*>self.ptr).AddBezier(&bz)

    def AddLine(self, float x, float y):
        cdef D2D1_POINT_2F pt
        pt.x = x
        pt.y = y
        (<ID2D1GeometrySink*>self.ptr).AddLine(pt)

    def AddQuadraticBezier(self, float x1, float y1, float x2, float y2):
        cdef D2D1_QUADRATIC_BEZIER_SEGMENT bz
        bz.point1.x = x1
        bz.point1.y = y1
        bz.point2.x = x2
        bz.point2.y = y2
        (<ID2D1GeometrySink*>self.ptr).AddQuadraticBezier(&bz)


cdef class Image(Resource):
    pass


cdef class Bitmap(Image):
    pass


_dwrite_factory = None

def GetDWriteFactory():
    global _dwrite_factory
    if _dwrite_factory is None:
        _dwrite_factory = DWriteFactory()
    return _dwrite_factory

class DirectWriteError(COMError):
    pass

cdef class DWriteFactory(COMObject):
    def __init__(self, int factoryType=0):
        cdef IDWriteFactory* factory
        res = DWriteCreateFactory(
            <DWRITE_FACTORY_TYPE>factoryType,
            IID_IDWriteFactory,
            <IUnknown**>&factory)
        if FAILED(res):
            raise DirectWriteError(res)
        COMObject.__init__(self, <int><intptr_t>factory)

    def CreateTextFormat(self, str familyName, float size, int weight=500, int style=0, int stretch=5):
        cdef IDWriteTextFormat *fmt
        cdef wchar_t *familyBuf
        familyBuf = PyUnicode_AsWideCharString(familyName, NULL)
        if familyBuf == NULL:
            raise MemoryError
        cdef wchar_t[1] localeBuf
        localeBuf[0] = 0
        res = (<IDWriteFactory*>self.ptr).CreateTextFormat(
            familyBuf,
            NULL,
            <DWRITE_FONT_WEIGHT>weight,
            <DWRITE_FONT_STYLE>style,
            <DWRITE_FONT_STRETCH>stretch,
            size,
            localeBuf,
            <IDWriteTextFormat**>&fmt)
        PyMem_Free(<void*>familyBuf)
        if FAILED(res):
            raise DirectWriteError(res)
        return TextFormat(<int><intptr_t>fmt)

    def CreateTextLayout(self, str text, TextFormat textFormat, float maxWidth, float maxHeight):
        cdef IDWriteTextLayout* layout
        cdef wchar_t *stringBuf
        cdef Py_ssize_t stringBufLen
        stringBuf = PyUnicode_AsWideCharString(text, &stringBufLen)
        if stringBuf == NULL:
            raise MemoryError
        res = (<IDWriteFactory*>self.ptr).CreateTextLayout(
            stringBuf,
            <UINT32>stringBufLen,
            <IDWriteTextFormat*>textFormat.ptr,
            maxWidth,
            maxHeight,
            &layout)
        PyMem_Free(stringBuf)
        if FAILED(res):
            raise DirectWriteError(res)
        return TextLayout(<int><intptr_t>layout)


cdef class FontFace(COMObject):
    pass


cdef class TextFormat(COMObject):
    pass


cdef class TextLayout(TextFormat):
    def GetMaxWidth(self):
        return (<IDWriteTextLayout*>self.ptr).GetMaxWidth()

    def GetMetrics(self):
        cdef DWRITE_TEXT_METRICS tm
        res = (<IDWriteTextLayout*>self.ptr).GetMetrics(&tm)
        if FAILED(res):
            raise DirectWriteError(res)
        TM = TEXT_METRICS()
        TM.left = tm.left
        TM.top = tm.top
        TM.width = tm.width
        TM.widthIncludingTrailingWhitespace = tm.widthIncludingTrailingWhitespace
        TM.height = tm.height
        TM.layoutWidth = tm.layoutWidth
        TM.layoutHeight = tm.layoutHeight
        TM.maxBidiReorderingDepth = tm.maxBidiReorderingDepth
        TM.lineCount = tm.lineCount
        return TM

class TEXT_METRICS:
    pass


"""
pyd2d - A Python wrapper for the Windows Direct2D and DirectWrite APIs.

This module provides a Pythonic interface to Direct2D and DirectWrite,
which are part of the Windows API.
"""

from typing import Optional, Tuple

from _pyd2d_const import *  # noqa: F403

class COMError(OSError):
    """
    Base class for COM errors.
    """

    hresult: int
    def __init__(self, hresult: int) -> None: ...

class Direct2DError(COMError):
    """
    Base class for Direct2D errors.
    """

def InitializeCOM(options: int = 0) -> None:
    """
    Initializes COM for the current thread by calling CoInitializeEx.
    """

class COMObject:
    """
    Base class for COM objects.
    Use the factory functions to create instances of COMObject subclasses.
    """

def GetD2DFactory() -> "D2DFactory":
    """
    Returns the Direct2D factory object, creating it on the first call.
    """

class D2DFactory(COMObject):
    """
    Creates Direct2D resources.
    """
    def __init__(self, factoryType: int = 0, debugLevel: int = 0) -> None:
        """
        Creates the Direct2D factory object.
        """
    def CreateHwndRenderTarget(
        self,
        hwnd: int,
        width: int,
        height: int,
        presentOptions: int = 0,
        rtType: int = 0,
        pixelFormat: int = 0,
        alphaMode: int = 0,
        dpiX: float = 0,
        dpiY: float = 0,
        usage: int = 0,
        featureLevel: int = 0,
    ) -> "HWNDRenderTarget":
        """
        Creates a HwndRenderTarget.
        """
    def CreatePathGeometry(self) -> "PathGeometry":
        """
        Creates a PathGeometry.
        """
    def CreateStrokeStyle(
        self,
        startCap: int = 0,
        endCap: int = 0,
        dashCap: int = 0,
        lineJoin: int = 0,
        miterLimit: float = 10.0,
        dashStyle: int = 0,
        dashOffset: float = 0.0,
    ) -> "StrokeStyle":
        """
        Creates a StrokeStyle.
        """

class Resource(COMObject):
    """
    Base class representing a Direct2D drawing resource.
    """

class RenderTarget(Resource):
    """
    Base class representing an object that can receive drawing commands.

    Classes that inherit from RenderTarget render the drawing commands they receive in different ways.
    """
    def BeginDraw(self) -> None:
        """
        Initiates drawing on this render target.
        """
    def Clear(self, r: float, g: float, b: float, a: float = 1.0) -> None:
        """
        Clears the drawing area to the specified color.
        """
    def CreateSolidColorBrush(
        self, r: float, g: float, b: float, a: float = 1.0, opacity: float = 1.0
    ) -> "SolidColorBrush":
        """
        Creates a SolidColorBrush.
        """
    def DrawBitmap(
        self,
        bitmap: "Bitmap",
        l: float,
        t: float,
        r: float,
        b: float,
        opacity: float = 1.0,
        interpolationMode: int = 1,
        srcRect: Optional[Tuple[float, float, float, float]] = None,
    ) -> None:
        """
        Draws the specified bitmap after scaling it to the size of the specified rectangle.
        """
    def DrawEllipse(
        self,
        cx: float,
        cy: float,
        rx: float,
        ry: float,
        brush: "Brush",
        strokeWidth: float = 1.0,
        strokeStyle: Optional["StrokeStyle"] = None,
    ) -> None:
        """
        Draws the outline of the specified ellipse using the specified stroke style.
        """
    def DrawGeometry(
        self,
        geometry: "Geometry",
        brush: "Brush",
        strokeWidth: float = 1.0,
        strokeStyle: Optional["StrokeStyle"] = None,
    ) -> None:
        """
        Draws the outline of the specified geometry using the specified stroke style.
        """
    def DrawLine(
        self,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        brush: "Brush",
        strokeWidth: float = 1.0,
        strokeStyle: Optional["StrokeStyle"] = None,
    ) -> None:
        """
        Draws a line between the specified points using the specified stroke style.
        """
    def DrawRectangle(
        self,
        l: float,
        t: float,
        r: float,
        b: float,
        brush: "Brush",
        strokeWidth: float = 1.0,
        strokeStyle: Optional["StrokeStyle"] = None,
    ) -> None:
        """
        Draws the outline of a rectangle that has the specified dimensions and stroke style.
        """
    def DrawText(
        self,
        text: str,
        textFormat: "TextFormat",
        l: float,
        t: float,
        r: float,
        b: float,
        brush: "Brush",
        options: int = 0,
        measuringMode: int = 0,
    ) -> None:
        """
        Draws the specified text using the format information provided by a TextFormat object.
        """
    def DrawTextLayout(
        self,
        x: float,
        y: float,
        textLayout: "TextLayout",
        brush: "Brush",
        options: int = 0,
    ) -> None:
        """
        Draws the formatted text described by the specified TextLayout object.
        """
    def EndDraw(self) -> None:
        """
        Ends drawing operations on the render target and indicates the current error state and associated tags.
        """
    def FillEllipse(
        self, cx: float, cy: float, rx: float, ry: float, brush: "Brush"
    ) -> None:
        """
        Paints the interior of the specified ellipse.
        """
    def FillGeometry(self, geometry: "Geometry", brush: "Brush") -> None:
        """
        Paints the interior of the specified geometry.
        """
    def FillRectangle(
        self, l: float, t: float, r: float, b: float, brush: "Brush"
    ) -> None:
        """
        Paints the interior of the specified rectangle.
        """
    def GetTransform(self) -> Tuple[float, float, float, float, float, float]:
        """
        Gets the current transform of the render target.
        """
    def SetAntialiasMode(self, mode: int) -> None:
        """
        Sets the antialiasing mode of the render target.
        The antialiasing mode applies to all subsequent drawing operations,
        excluding text and glyph drawing operations.
        """
    def SetTransform(
        self,
        m11: float = 1,
        m12: float = 0,
        m21: float = 0,
        m22: float = 1,
        dx: float = 0,
        dy: float = 0,
    ) -> None:
        """
        Applies the specified transform to the render target,
        replacing the existing transformation.
        All subsequent drawing operations occur in the transformed space.
        """

class HWNDRenderTarget(RenderTarget):
    """
    Renders drawing instructions to a window.
    """
    def Resize(self, width: int, height: int) -> None:
        """
        Changes the size of the render target to the specified pixel size.
        """

class Brush(Resource):
    """
    Base class representing an object that paints an area.
    Classes that derive from Brush describe how the area is painted.
    """
    def GetOpacity(self) -> float:
        """
        Gets the degree of opacity of this brush.
        """

class SolidColorBrush(Brush):
    """
    Paints an area with a solid color.
    """
    def SetColor(self, r: float, g: float, b: float, a: float = 1.0) -> None:
        """
        Specifies the color of this solid color brush.
        """

class StrokeStyle(Resource):
    """
    Describes the caps, miter limit, line join, and dash information for a stroke.
    """

class Geometry(Resource):
    """
    Base class representing a geometry resource.
    Defines a set of helper methods for manipulating and measuring geometric shapes.
    Classes that inherit from Geometry define specific shapes.
    """

class PathGeometry(Geometry):
    """
    Represents a complex shape that may be composed of arcs, curves, and lines.
    """
    def Open(self) -> "GeometrySink":
        """
        Retrieves the GeometrySink that is used to populate the path geometry with figures and segments.
        """

class SimplifiedGeometrySink(COMObject):
    """
    Describes a geometric path that does not contain quadratic bezier curves or arcs.
    """
    def BeginFigure(self, x: float, y: float, figureBegin: int = 0) -> None:
        """
        Starts a new figure at the specified point.
        """
    def Close(self) -> None:
        """
        Closes the geometry sink, indicates whether it is in an error state, and resets the sink's error state.
        """
    def EndFigure(self, figureEnd: int = 0) -> None:
        """
        Ends the current figure; optionally, closes it.
        """
    def SetFillMode(self, fillMode: int) -> None:
        """
        Specifies the method used to determine which pointsare inside
        the geometry described by this geometry sink and which points are outside.
        """

class GeometrySink(SimplifiedGeometrySink):
    """
    Describes a geometric path that can contain lines, arcs,
    cubic Bezier curves, and quadratic Bezier curves.
    """
    def AddArc(
        self,
        x: float,
        y: float,
        rx: float,
        ry: float,
        rotationAngle: float,
        sweepDirection: int,
        arcSize: int,
    ) -> None:
        """
        Adds a single arc to the path geometry.
        """
    def AddBezier(
        self, x1: float, y1: float, x2: float, y2: float, x3: float, y3: float
    ) -> None:
        """
        Creates a cubic Bezier curve between the current point and the specified end point.
        """
    def AddLine(self, x: float, y: float) -> None:
        """
        Creates a line segment between the current point and the specified end point
        and adds it to the geometry sink.
        """
    def AddQuadraticBezier(self, x1: float, y1: float, x2: float, y2: float) -> None:
        """
        Creates a quadratic Bezier curve between the current point and the specified end point.
        """

class Image(Resource):
    """
    Base class representing a producer of pixels that can fill an arbitrary 2D plane.
    """

class Bitmap(Image):
    """
    Represents a bitmap that has been bound to a RenderTarget.
    """

def GetDWriteFactory() -> "DWriteFactory":
    """
    Returns the DirectWrite factory object, creating it on the first call.
    """

class DirectWriteError(COMError):
    """
    Base class for DirectWrite errors.
    """

class DWriteFactory(COMObject):
    """
    Creates DirectWrite resources.
    """
    def __init__(self, factoryType: int = 0) -> None:
        """
        Creates the DirectWrite factory object.
        """
    def CreateTextFormat(
        self,
        familyName: str,
        size: float,
        weight: int = 500,
        style: int = 0,
        stretch: int = 5,
    ) -> "TextFormat":
        """
        Creates a text format object used for text layout.
        """
    def CreateTextLayout(
        self, text: str, textFormat: "TextFormat", maxWidth: float, maxHeight: float
    ) -> "TextLayout":
        """
        Takes a string, text format, and associated constraints,
        and produces an object that represents the fully analyzed and formatted result.
        """

class FontFace(COMObject):
    """
    Exposes various font data such as metrics, names, and glyph outlines.
    Contains font face type, appropriate file references, and face identification data.
    """

class TextFormat(COMObject):
    """
    Describes the font and paragraph properties used to format text,
    and describes locale information.
    """

class TextLayout(TextFormat):
    """
    Represents a block of text after it has been fully analyzed and formatted.
    """
    def GetMaxWidth(self) -> float:
        """
        Gets the layout maximum width.
        """
    def GetMetrics(self) -> "TEXT_METRICS":
        """
        Retrieves overall metrics for the formatted string.
        """

class TEXT_METRICS:
    """
    Contains the metrics associated with text after layout.
    All coordinates are in device independent pixels (DIPs).
    """

    left: float
    top: float
    width: float
    widthIncludingTrailingWhitespace: float
    height: float
    layoutWidth: float
    layoutHeight: float
    maxBidiReorderingDepth: int
    lineCount: int
